package org.softme.webpm.test;
import java.io.*;
import java.util.*;

/**
 * Syntax highlights java by turning it into html.
 *
 * A codeviewer object is created and then keeps state as lines are passed
 * in. Each line passed in as java test, is returned as syntax highlighted
 * html text.
 *
 * Users of the class can set how the java code will be highlighted with setter
 * methods.
 *
 * Only valid java lines should be passed in since the object maintains state
 * and may not handle illegal code gracefully.
 *
 * The actual system is implemented as a series of filters that deal with
 * specific portions of the java code. The filters are as follows:
 *
 *  htmlFilter
 *     |__
 *        multiLineCommentFilter
 *           |___
 *                inlineCommentFilter
 *                   |___
 *                        stringFilter
 *                           |__
 *                               keywordFilter
 */
public class CodeTransform implements Serializable {

    public static final boolean LINE_NUMBERS = true;

    private BufferedReader in;
    private StringBuffer out;
    //private HashMap reservedWords = new HashMap();
    private Hashtable reservedWords = new Hashtable();
    private boolean inMultiLineComment = false;
    private String backgroundColor = "#ffffff";
    private String commentStart = "<font color=\"#aa0000\"><i>";
    private String commentEnd = "</font></i>";
    private String stringStart = "<font color=\"#000099\">";
    private String stringEnd = "</font>";
    private String reservedWordStart = "<b>";
    private String reservedWordEnd = "</b>";

    public CodeTransform() {
        loadHash();
    }

    public void setCommentStart(String commentStart) {
        this.commentStart = commentStart;
    }
    public void setCommentEnd(String commentEnd) {
        this.commentEnd = commentEnd;
    }
    public void setStringStart(String stringStart) {
        this.stringStart = stringStart;
    }
    public void setStringEnd(String stringEnd) {
        this.stringEnd = stringEnd;
    }
    public void setReservedWordStart(String reservedWordStart) {
        this.reservedWordStart = reservedWordStart;
    }
    public void setReservedWordEnd(String reservedWordEnd) {
        this.reservedWordEnd = reservedWordEnd;
    }

    public String getCommentStart() { return commentStart; }
    public String getCommentEnd() { return commentEnd; }
    public String getStringStart() { return stringStart; }
    public String getStringEnd() { return stringEnd; }
    public String getReservedWordStart() { return reservedWordStart; }
    public String getReservedWordEnd() { return reservedWordEnd; }

    /**
     * Now different method of seeing if at end of input stream,
     * closes inputs stream at end.
     */
    public String syntaxHighlight(String line) {
       return htmlFilter(line);
    }

    /**
     * Filter html tags into more benign text.
     */
    private String htmlFilter( String line ) {
        StringBuffer buf = new StringBuffer();
        if( line == null || line.equals("") ) {
            return "";
        }
        //line = replace(line,"<","&lt;");
       // line = replace(line,">","&gt;");
        line = replace( line, "\\\\", "&#47;&#47;" );
        line = replace( line, "\\\"", "\\&quot;" );
        line = replace( line, "'\"'", "'&quot;'" );
        line = replace(line," ","&nbsp;");
        line = replace(line,"\n","<br>");
        return multiLineCommentFilter(line);
    }

    /**
     * Filter out multiLine comments. State is kept with a private boolean
     * variable.
     */
    private String multiLineCommentFilter(String line) {
        if (line == null || line.equals("")) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        int index;
        //First, check for the end of a multi-line comment.
        if (inMultiLineComment && (index = line.indexOf("*/")) > -1 && !isInsideString(line,index)) {
            inMultiLineComment = false;
            buf.append(line.substring(0,index));
            buf.append("*/").append(commentEnd);
            if (line.length() > index+2) {
                buf.append(inlineCommentFilter(line.substring(index+2)));
            }
            return buf.toString();
        }
        //If there was no end detected and we're currently in a multi-line
        //comment, we don't want to do anymore work, so return line.
        else if (inMultiLineComment) {
            return line;
        }
        //We're not currently in a comment, so check to see if the start
        //of a multi-line comment is in this line.
        else if ((index = line.indexOf("/*")) > -1 && !isInsideString(line,index)) {
            inMultiLineComment = true;
            //Return result of other filters + everything after the start
            //of the multiline comment. We need to pass the through the
            //to the multiLineComment filter again in case the comment ends
            //on the same line.
            buf.append(inlineCommentFilter(line.substring(0,index)));
            buf.append(commentStart).append("/*");
            buf.append(multiLineCommentFilter(line.substring(index+2)));
            return buf.toString();
        }
        //Otherwise, no useful multi-line comment information was found so
        //pass the line down to the next filter for processesing.
        else {
            return inlineCommentFilter(line);
        }
    }

    /**
     * Filter inline comments from a line and formats them properly.
     *
     * One problem we'll have to solve here: comments contained in a string
     * should be ignored... this is also true of the multiline comments. So,
     * we could either ignore the problem, or implement a function called
     * something like isInsideString(line, index) where index points to
     * some point in the line that we need to check... started doing this
     * function below.
     */
    private String inlineCommentFilter(String line) {
        if (line == null || line.equals("")) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        int index;
        if ((index = line.indexOf("//")) > -1 && !isInsideString(line,index)) {
            buf.append(stringFilter(line.substring(0,index)));
            buf.append(commentStart);
            buf.append(line.substring(index));
            buf.append(commentEnd);
        }
        else {
            buf.append(stringFilter(line));
        }
        return buf.toString();
    }

    /**
     * Filters strings from a line of text and formats them properly.
     */
    private String stringFilter(String line) {
        if (line == null || line.equals("")) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        if (line.indexOf("\"") <= -1) {
            return keywordFilter(line);
        }
        int start = 0;
        int startStringIndex = -1;
        int endStringIndex = -1;
        int tempIndex;
        //Keep moving through String characters until we want to stop...
        while ((tempIndex = line.indexOf("\"")) > -1) {
            //We found the beginning of a string
            if (startStringIndex == -1) {
                startStringIndex = 0;
                buf.append( stringFilter(line.substring(start,tempIndex)) );
                buf.append(stringStart).append("\"");
                line = line.substring(tempIndex+1);
            }
            //Must be at the end
            else {
                startStringIndex = -1;
                endStringIndex = tempIndex;
                buf.append(line.substring(0,endStringIndex+1));
                buf.append(stringEnd);
                line = line.substring(endStringIndex+1);
            }
        }

        buf.append( keywordFilter(line) );

        return buf.toString();
    }

    /**
     * Filters keywords from a line of text and formats them properly.
     */
    private String keywordFilter( String line ) {
        if( line == null || line.equals("") ) {
            return "";
        }
        StringBuffer buf = new StringBuffer();
        //HashMap usedReservedWords = new HashMap();
        Hashtable usedReservedWords = new Hashtable();
        int i=0, startAt=0;
        char ch;
        StringBuffer temp = new StringBuffer();
        while( i < line.length() ) {
            temp.setLength(0);
            ch = line.charAt(i);
            startAt = i;
            while( i<line.length() && ( ( ch >= 65 && ch <= 90 )
                    || ( ch >= 97 && ch <= 122 ) ) ) {
                temp.append(ch);
                i++;
                if( i < line.length() ) {
                    ch = line.charAt(i);
                }
            }
            String tempString = temp.toString();
            if( reservedWords.containsKey(tempString) && !usedReservedWords.containsKey(tempString)) {
                usedReservedWords.put(tempString,tempString);
                line = replace( line, tempString, (reservedWordStart+tempString+reservedWordEnd) );
                i += (reservedWordStart.length() + reservedWordEnd.length());
            }
            else {
                i++;
            }
        }
        buf.append(line);
        return buf.toString();
    }

    /**
     *  Replace...
     *
     *  I made it use a stringBuffer... hope it still works :)
     */
    private String replace( String line, String oldString, String newString ) {
        int i=0;
        while( ( i=line.indexOf( oldString, i ) ) >= 0 ) {
            line = (new StringBuffer().append(line.substring(0,i)).append(newString).append(line.substring(i+oldString.length()))).toString();
            i += newString.length();
        }
        return line;
    }

    /**
     * Checks to see if some position in a line is between String start and
     * ending characters. Not yet used in code or fully working :)
     */
    private boolean isInsideString(String line, int position) {
        if (line.indexOf("\"") < 0) {
            return false;
        }
        int index;
        String left = line.substring(0,position);
        String right = line.substring(position);
        int leftCount = 0;
        int rightCount = 0;
        while ((index = left.indexOf("\"")) > -1) {
            leftCount ++;
            left = left.substring(index+1);
        }
        while ((index = right.indexOf("\"")) > -1) {
            rightCount ++;
            right = right.substring(index+1);
        }
        if (rightCount % 2 != 0 && leftCount % 2 != 0) {
            return true;
        }
        else {
            return false;
        }
    }

    private void loadHash() {
        reservedWords.put( "abstract", "abstract" );
        reservedWords.put( "do", "do" );
        reservedWords.put( "inner", "inner" );
        reservedWords.put( "public", "public" );
        reservedWords.put( "var", "var" );
        reservedWords.put( "boolean", "boolean" );
        reservedWords.put( "continue", "continue" );
        reservedWords.put( "int", "int" );
        reservedWords.put( "return", "return" );
        reservedWords.put( "void", "void" );
        reservedWords.put( "break", "break" );
        reservedWords.put( "else", "else" );
        reservedWords.put( "interface", "interface" );
        reservedWords.put( "short", "short" );
        reservedWords.put( "volatile", "volatile" );
        reservedWords.put( "byvalue", "byvalue" );
        reservedWords.put( "extends", "extends" );
        reservedWords.put( "long", "long" );
        reservedWords.put( "static", "static" );
        reservedWords.put( "while", "while" );
        reservedWords.put( "case", "case" );
        reservedWords.put( "final", "final" );
        reservedWords.put( "naive", "naive" );
        reservedWords.put( "super", "super" );
        reservedWords.put( "transient", "transient" );
        reservedWords.put( "cast", "cast" );
        reservedWords.put( "float", "float" );
        reservedWords.put( "new", "new" );
        reservedWords.put( "rest", "rest" );
        reservedWords.put( "catch", "catch" );
        reservedWords.put( "for", "for" );
        reservedWords.put( "null", "null" );
        reservedWords.put( "synchronized", "synchronized" );
        reservedWords.put( "char", "char" );
        reservedWords.put( "finally", "finally" );
        reservedWords.put( "operator", "operator" );
        reservedWords.put( "this", "this" );
        reservedWords.put( "class", "class" );
        reservedWords.put( "generic", "generic" );
        reservedWords.put( "outer", "outer" );
        reservedWords.put( "switch", "switch" );
        reservedWords.put( "const", "const" );
        reservedWords.put( "goto", "goto" );
        reservedWords.put( "package", "package" );
        reservedWords.put( "throw", "throw" );
        reservedWords.put( "double", "double" );
        reservedWords.put( "if", "if" );
        reservedWords.put( "private", "private" );
        reservedWords.put( "true", "true" );
        reservedWords.put( "default", "default" );
        reservedWords.put( "import", "import" );
        reservedWords.put( "protected", "protected" );
        reservedWords.put( "try", "try" );
    }

    void writeObject(ObjectOutputStream oos) throws IOException {
        oos.defaultWriteObject();
    }

    void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
        ois.defaultReadObject();
    }

    public String getHeader() {

        //FIX ME: migrate this to use ECS
        return "<html>\n" +
               "<body bgcolor=\"white\">\n" +
               "<pre>\n";
    }

    public String getFooter() {

        //FIX ME: migrate this to use ECS
        return "</pre>\n" +
               "<hr>" +
               "<center>" +"JWIKI"+ "</center>" +
               "</body>\n" +
               "</html>\n";
    }


    public  String transform(String line){
        int linenumber = 1;
        String row="<a name=\"" +  linenumber + "\" " +
        "href=\"#" + linenumber + "\">" +
        linenumber +
        "</a>  " ;
        line=this.syntaxHighlight( line ) ;
        line=row+line;
        ++linenumber;
        return line;
    }
    /**
    This is the public method for doing all transforms of code.

    */
    public void transform( String sourcefile, String destfile )
        throws IOException
    {



        //make sure that the parent directories exist...
        new File( new File( destfile).getParent() ).mkdirs();


        BufferedReader in = new BufferedReader( new FileReader( sourcefile ) );;
        PrintWriter out = new PrintWriter( new FileWriter( destfile ) );

        String line = "";

        out.println( getHeader() );
        int linenumber = 1;
        while ( (line = in.readLine()) != null ) {
           if ( LINE_NUMBERS ) {
                out.print( "<a name=\"" +  linenumber + "\" " +
                           "href=\"#" + linenumber + "\">" +
                           linenumber +
                           "</a>  " );
           }

           out.println( this.syntaxHighlight( line ) );

           ++linenumber;
        }

        out.println( getFooter() );
        out.flush();

    }

}
